"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""
from typing import Optional

from django.contrib import messages
from django.core.exceptions import ValidationError, ImproperlyConfigured
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import (
    ArchiveIndexView, YearArchiveView,
    MonthArchiveView, DetailView,
    RedirectView, CreateView,
    DeleteView, UpdateView
)
from django.views.generic.detail import SingleObjectMixin

from django_ledger.forms.closing_entry import ClosingEntryCreateForm, ClosingEntryUpdateForm
from django_ledger.io.io_core import get_localdate
from django_ledger.models import EstimateModel, ClosingEntryModelQuerySet
from django_ledger.models.closing_entry import ClosingEntryModel
from django_ledger.models.entity import EntityModel
from django_ledger.views import DjangoLedgerSecurityMixIn


class ClosingEntryModelBaseView(DjangoLedgerSecurityMixIn):
    queryset: Optional[ClosingEntryModelQuerySet] = None

    def get_queryset(self):
        if self.queryset is None:
            entity_model: EntityModel = self.get_authorized_entity_instance()
            closing_entry_model_qs = entity_model.closingentrymodel_set.all().select_related(
                'entity_model',
                'ledger_model'
            ).order_by('-closing_date')
            self.queryset = closing_entry_model_qs
        return self.queryset


class ClosingEntryModelListView(ClosingEntryModelBaseView, ArchiveIndexView):
    template_name = 'django_ledger/closing_entry/closing_entry_list.html'
    date_field = 'closing_date'
    allow_future = False
    context_object_name = 'closing_entry_list'
    PAGE_TITLE = _('Closing Entry List')
    paginate_by = 10
    paginate_orphans = 2
    allow_empty = True
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE,
        'header_subtitle_icon': 'file-icons:finaldraft'
    }


class ClosingEntryModelYearListView(ClosingEntryModelListView, YearArchiveView):
    paginate_by = 10
    make_object_list = True


class ClosingEntryModelMonthListView(ClosingEntryModelListView, MonthArchiveView):
    paginate_by = 10
    month_format = '%m'
    date_list_period = 'year'


class ClosingEntryModelCreateView(ClosingEntryModelBaseView, CreateView):
    template_name = 'django_ledger/closing_entry/closing_entry_create.html'
    PAGE_TITLE = _('Create Closing Entry')
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE,
        'header_subtitle_icon': 'file-icons:finaldraft'
    }
    form_class = ClosingEntryCreateForm

    def get_initial(self):
        return {
            'closing_date': get_localdate(),
            'entity_model': self.AUTHORIZED_ENTITY_MODEL
        }

    def get_form_kwargs(self, **kwargs):
        return super().get_form_kwargs(**kwargs) | {
            'entity_model': self.AUTHORIZED_ENTITY_MODEL
        }

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        entity_model: EntityModel = self.get_authorized_entity_instance()
        ctx['header_subtitle'] = entity_model.name
        return ctx

    def form_valid(self, form):
        closing_date = form.cleaned_data['closing_date']
        entity_model: EntityModel = self.get_authorized_entity_instance()
        ce_model, _ = entity_model.close_entity_books(
            closing_date=closing_date,
            force_update=True,
            post_closing_entry=False
        )
        self.ce_model: EstimateModel = ce_model
        return HttpResponseRedirect(self.get_success_url())

    def get_success_url(self):
        return reverse(
            viewname='django_ledger:closing-entry-detail',
            kwargs={
                'entity_slug': self.kwargs['entity_slug'],
                'closing_entry_pk': self.ce_model.uuid
            }
        )


class ClosingEntryModelDetailView(ClosingEntryModelBaseView, DetailView):
    template_name = 'django_ledger/closing_entry/closing_entry_detail.html'
    pk_url_kwarg = 'closing_entry_pk'
    context_object_name = 'closing_entry_model'
    extra_context = {
        'header_title': 'Closing Entry Detail',
        'header_subtitle_icon': 'file-icons:finaldraft'
    }

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        closing_entry_model = ctx['object']
        ctx['page_title'] = f'Closing Entry {closing_entry_model.closing_date}'
        closing_entry_txs_qs = closing_entry_model.closingentrytransactionmodel_set.all()
        ctx['closing_entry_txs_qs'] = closing_entry_txs_qs.select_related(
            'account_model', 'account_model__coa_model', 'unit_model')
        return ctx


class ClosingEntryModelUpdateView(ClosingEntryModelBaseView, UpdateView):
    template_name = 'django_ledger/closing_entry/closing_entry_update.html'
    pk_url_kwarg = 'closing_entry_pk'
    form_class = ClosingEntryUpdateForm
    context_object_name = 'closing_entry_model'
    extra_context = {
        'header_title': 'Closing Entry Detail',
        'header_subtitle_icon': 'file-icons:finaldraft'
    }

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        closing_entry_model = ctx['object']
        ctx['page_title'] = f'Closing Entry {closing_entry_model.closing_date} Update'
        closing_entry_txs_qs = closing_entry_model.closingentrytransactionmodel_set.all()
        ctx['closing_entry_txs_qs'] = closing_entry_txs_qs.select_related(
            'account_model', 'account_model__coa_model', 'unit_model')
        return ctx

    def get_success_url(self):
        return reverse(
            viewname='django_ledger:closing-entry-detail',
            kwargs={
                'entity_slug': self.kwargs['entity_slug'],
                'closing_entry_pk': self.object.uuid
            }
        )


class ClosingEntryDeleteView(ClosingEntryModelBaseView, DeleteView):
    template_name = 'django_ledger/closing_entry/closing_entry_delete.html'
    pk_url_kwarg = 'closing_entry_pk'
    context_object_name = 'closing_entry'
    extra_context = {
        'header_title': _('Delete Closing Entry'),
        'header_subtitle_icon': 'file-icons:finaldraft'
    }

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        closing_entry_model: ClosingEntryModel = self.object
        entity_model: EntityModel = self.get_authorized_entity_instance()
        context['page_title'] = 'Delete Closing Entry {}'.format(closing_entry_model.closing_date)
        context['header_subtitle'] = entity_model.name
        return context

    def get_success_url(self):
        return reverse(viewname='django_ledger:closing-entry-list',
                       kwargs={
                           'entity_slug': self.kwargs['entity_slug']
                       })


class ClosingEntryModelActionView(ClosingEntryModelBaseView,
                                  RedirectView,
                                  SingleObjectMixin):
    http_method_names = ['get']
    pk_url_kwarg = 'closing_entry_pk'
    action_name = None
    commit = True

    def get_redirect_url(self, *args, **kwargs):
        return reverse('django_ledger:closing-entry-detail',
                       kwargs={
                           'entity_slug': kwargs['entity_slug'],
                           'closing_entry_pk': kwargs['closing_entry_pk']
                       })

    def get(self, request, *args, **kwargs):
        kwargs['user_model'] = self.request.user
        if not self.action_name:
            raise ImproperlyConfigured('View attribute action_name is required.')
        response = super(ClosingEntryModelActionView, self).get(request, *args, **kwargs)
        closing_entry_model: ClosingEntryModel = self.get_object()

        try:
            getattr(closing_entry_model, self.action_name)(commit=self.commit, **kwargs)
        except ValidationError as e:
            messages.add_message(request,
                                 message=e.message,
                                 level=messages.ERROR,
                                 extra_tags='is-danger')
        return response



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.
"""
from typing import Optional

from django.contrib import messages
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.utils.translation import gettext_lazy as _
from django.views.generic import UpdateView, ListView, RedirectView, CreateView
from django.views.generic.detail import SingleObjectMixin

from django_ledger.forms.chart_of_accounts import ChartOfAccountsModelUpdateForm, ChartOfAccountsModelCreateForm
from django_ledger.models import EntityModel, ChartOfAccountModelQuerySet
from django_ledger.models.chart_of_accounts import ChartOfAccountModel
from django_ledger.views.mixins import DjangoLedgerSecurityMixIn


class ChartOfAccountModelModelBaseViewMixIn(DjangoLedgerSecurityMixIn):
    queryset: Optional[ChartOfAccountModelQuerySet] = None

    def get_queryset(self):
        if self.queryset is None:
            entity_model: EntityModel = self.get_authorized_entity_instance()
            self.queryset: ChartOfAccountModelQuerySet = entity_model.chartofaccountmodel_set.all().order_by('-updated')
        return self.queryset


class ChartOfAccountModelListView(ChartOfAccountModelModelBaseViewMixIn, ListView):
    template_name = 'django_ledger/chart_of_accounts/coa_list.html'
    context_object_name = 'coa_list'
    inactive = False

    def get_queryset(self):
        qs = super().get_queryset()
        if self.inactive:
            return qs.filter(active=False)
        return qs.active()

    def get_context_data(self, *, object_list=None, **kwargs):
        context = super().get_context_data(object_list=None, **kwargs)
        context['inactive'] = self.inactive
        context['header_subtitle'] = self.AUTHORIZED_ENTITY_MODEL.name
        context['header_subtitle_icon'] = 'gravity-ui:hierarchy'
        context['page_title'] = 'Inactive Chart of Account List' if self.inactive else 'Chart of Accounts List'
        context['header_title'] = 'Inactive Chart of Account List' if self.inactive else 'Chart of Accounts List'
        return context


class ChartOfAccountModelCreateView(ChartOfAccountModelModelBaseViewMixIn, CreateView):
    template_name = 'django_ledger/chart_of_accounts/coa_create.html'
    extra_context = {
        'header_title': _('Create Chart of Accounts'),
        'page_title': _('Create Chart of Account'),
    }

    def get_initial(self):
        return {
            'entity': self.get_authorized_entity_instance(),
        }

    def get_form(self, form_class=None):
        return ChartOfAccountsModelCreateForm(
            entity_model=self.get_authorized_entity_instance(),
            **self.get_form_kwargs()
        )

    def get_context_data(self, *, object_list=None, **kwargs):
        context = super().get_context_data(object_list=None, **kwargs)
        context['header_subtitle'] = f'New Chart of Accounts: {self.AUTHORIZED_ENTITY_MODEL.name}'
        context['header_subtitle_icon'] = 'gravity-ui:hierarchy'
        return context

    def get_success_url(self):
        chart_of_accounts_model: ChartOfAccountModel = self.object
        return chart_of_accounts_model.get_coa_list_url()


class ChartOfAccountModelUpdateView(ChartOfAccountModelModelBaseViewMixIn, UpdateView):
    context_object_name = 'coa_model'
    slug_url_kwarg = 'coa_slug'
    template_name = 'django_ledger/chart_of_accounts/coa_update.html'
    form_class = ChartOfAccountsModelUpdateForm

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        chart_of_accounts_model: ChartOfAccountModel = self.object
        context['page_title'] = f'Update Chart of Account {chart_of_accounts_model.name}'
        context['header_title'] = f'Update Chart of Account {chart_of_accounts_model.name}'
        return context

    def get_success_url(self):
        chart_of_accounts_model: ChartOfAccountModel = self.object
        return chart_of_accounts_model.get_coa_list_url()


# todo: centralize this functionality into a separate class for ALL Action views...
class CharOfAccountModelActionView(ChartOfAccountModelModelBaseViewMixIn,
                                   RedirectView,
                                   SingleObjectMixin):
    http_method_names = ['get']
    slug_url_kwarg = 'coa_slug'
    action_name = None
    commit = True

    def get_redirect_url(self, *args, **kwargs):
        chart_of_accounts_model: ChartOfAccountModel = self.get_object()
        return chart_of_accounts_model.get_coa_list_url()

    def get(self, request, *args, **kwargs):
        kwargs['user_model'] = self.request.user
        if not self.action_name:
            raise ImproperlyConfigured('View attribute action_name is required.')
        response = super(CharOfAccountModelActionView, self).get(request, *args, **kwargs)
        coa_model: ChartOfAccountModel = self.get_object()

        try:
            getattr(coa_model, self.action_name)(commit=self.commit, **kwargs)
            messages.add_message(request, level=messages.SUCCESS, extra_tags='is-success',
                                 message=_('Successfully updated {} Default Chart of Account to '.format(
                                     self.AUTHORIZED_ENTITY_MODEL.name) +
                                           '{}'.format(coa_model.name)))
        except ValidationError as e:
            messages.add_message(request,
                                 message=e.message,
                                 level=messages.ERROR,
                                 extra_tags='is-danger')
        return response



from datetime import datetime, date
from typing import Optional, Dict, Union

from django_ledger.io import IODigestContextManager
from django_ledger.report.core import BaseReportSupport, PDFReportValidationError
from django_ledger.settings import DJANGO_LEDGER_CURRENCY_SYMBOL
from django_ledger.templatetags.django_ledger import currency_symbol, currency_format


class BalanceSheetReport(BaseReportSupport):

    def __init__(self, *args, io_digest: IODigestContextManager, report_subtitle: Optional[str] = None, **kwargs):

        if not io_digest.has_balance_sheet():
            raise PDFReportValidationError('IO Digest does not have balance sheet information.')
        super().__init__(*args, io_digest=io_digest, report_subtitle=report_subtitle, **kwargs)
        self.TABLE_HEADERS = {
            'role': {
                'title': '',
                'spacing': 10,
                'align': 'L',
            },
            'code': {
                'title': 'Account Code',
                'spacing': 15,
                'align': 'L',
            },
            'name': {
                'title': 'Account Name',
                'spacing': 50,
                'align': 'C'
            },
            'balance_type': {
                'title': 'Balance Type',
                'spacing': 5,
                'align': 'C'
            },
            'balance': {
                'title': f'Balance ({DJANGO_LEDGER_CURRENCY_SYMBOL})',
                'spacing': 10,
                'align': 'R'
            },
            'total': {
                'title': f'Total ({DJANGO_LEDGER_CURRENCY_SYMBOL})',
                'spacing': 10,
                'align': 'R'
            }
        }
        for k, th in self.TABLE_HEADERS.items():
            th['width'] = self.get_string_width(th['title']) + th['spacing']

    def get_report_data(self) -> Dict:
        return self.IO_DIGEST.get_balance_sheet_data()

    def get_report_name(self) -> str:
        return 'Balance Sheet Statement'

    def print_section_data(self, section_data):

        for r, d in section_data['roles'].items():
            # # print role name...
            self.set_font(self.FONT_FAMILY, style='U', size=self.FONT_SIZE)
            self.set_x(15)
            self.cell(
                w=self.get_string_width(d['role_name']),
                h=2,
                txt=d['role_name']
            )
            self.ln(h=5)

            # for each account...
            self.set_font(self.FONT_FAMILY, size=self.FONT_SIZE)
            account_height = 3
            for acc in d['accounts']:
                self.cell(
                    w=self.TABLE_HEADERS['role']['width'],
                    h=account_height,
                    txt=''
                )
                self.cell(
                    w=self.TABLE_HEADERS['code']['width'],
                    h=account_height,
                    txt='      {code}'.format(code=acc['code'])
                )

                if acc['activity']:
                    act = ' '.join(acc['activity'].split('_')).lower()
                    self.cell(
                        w=self.TABLE_HEADERS['name']['width'],
                        h=account_height,
                        txt=f'{acc["name"]} ({act})'
                    )
                else:
                    self.cell(
                        w=self.TABLE_HEADERS['name']['width'],
                        h=account_height,
                        txt=acc['name']
                    )

                self.cell(
                    w=self.TABLE_HEADERS['balance_type']['width'],
                    h=account_height,
                    txt=acc['balance_type'],
                    align='C'
                )
                self.cell(
                    w=self.TABLE_HEADERS['balance']['width'],
                    h=account_height,
                    align='R',
                    txt=f"{self.CURRENCY_SYMBOL}{currency_format(acc['balance'])}"
                )
                self.ln(5)

            self.set_font(self.FONT_FAMILY, style='B', size=self.FONT_SIZE)
            self.cell(
                w=20,
                h=2,
                txt='Total {r}'.format(r=d['role_name'])
            )

            self.set_font(self.FONT_FAMILY, style='B', size=self.FONT_SIZE + 1)
            self.set_x(180)
            self.cell(
                w=20,
                h=2,
                align='R',
                txt='{s}{tot}'.format(s=currency_symbol(),
                                      tot=currency_format(d['total_balance']))
            )
            self.ln(3)
            self.print_hline()
            self.ln(5)

    def print_assets(self):

        self.print_section_title('Assets')
        self.ln(8)

        # accounts data....
        bs_data = self.IO_DIGEST.get_balance_sheet_data()
        section_data = bs_data.get('assets')

        if section_data:
            self.print_section_data(section_data)
            self.set_font(self.FONT_FAMILY, style='BU', size=self.FONT_SIZE + 3)
            self.set_x(180)
            self.cell(
                w=20,
                h=2,
                align='R',
                txt='Total Assets: {s}{tot}'.format(s=currency_symbol(),
                                                    tot=currency_format(section_data['total_balance']))
            )
        self.ln(h=5)

    def print_liabilities(self):
        self.print_section_title('Liabilities')
        self.ln(8)

        # accounts data....
        bs_data = self.IO_DIGEST.get_balance_sheet_data()
        section_data = bs_data.get('liabilities')
        if section_data:
            self.print_section_data(section_data)
            self.set_font(self.FONT_FAMILY, style='BU', size=self.FONT_SIZE + 1)
            self.set_x(180)
            self.cell(
                w=20,
                h=2,
                align='R',
                txt='Total Liabilities: {s}{tot}'.format(s=currency_symbol(),
                                                         tot=currency_format(section_data['total_balance']))
            )
        self.ln(h=5)

    def print_equity(self):

        self.print_section_title('Equity')
        self.ln(8)

        # accounts data....
        bs_data = self.IO_DIGEST.get_balance_sheet_data()
        section_data = bs_data.get('equity')

        if section_data:
            self.print_section_data(section_data)

        self.set_font(self.FONT_FAMILY, style='B', size=self.FONT_SIZE)
        self.cell(
            w=20,
            h=2,
            txt='Retained Earnings'
        )
        self.set_x(180)
        self.cell(
            w=20,
            h=2,
            align='R',
            txt='{s}{tot}'.format(r='Retained Earnings',
                                  s=currency_symbol(),
                                  tot=currency_format(bs_data['retained_earnings_balance']))
        )
        self.ln(h=3)
        self.print_hline()
        self.ln(5)

        self.set_font(self.FONT_FAMILY, style='BU', size=self.FONT_SIZE + 1)
        self.set_x(180)
        self.cell(
            w=20,
            h=2,
            align='R',
            txt='Total Equity: {s}{tot}'.format(s=currency_symbol(),
                                                tot=currency_format(bs_data['equity_balance']))
        )
        self.ln(h=10)

        self.set_font(self.FONT_FAMILY, style='BU', size=self.FONT_SIZE + 3)
        self.set_x(180)
        self.cell(
            w=20,
            h=2,
            align='R',
            txt='Total Liabilities + Equity: {s}{tot}'.format(s=currency_symbol(),
                                                              tot=currency_format(
                                                                  bs_data['liabilities_equity_balance']))
        )
        self.ln(h=5)

    def get_pdf_filename(self,
                         to_dt: Optional[Union[datetime, date]] = None,
                         dt_strfmt: str = '%Y%m%d'):
        if to_dt:
            to_dt = to_dt.strftime(dt_strfmt if dt_strfmt else self.IO_DIGEST.get_strftime_format())
        else:
            to_dt = self.IO_DIGEST.get_to_datetime(fmt=dt_strfmt, as_str=True)
        f_name = f'{self.get_report_title()}_BalanceSheetStatement_{to_dt}.pdf'
        return f_name

    def create_pdf_report(self):
        self.print_headers()
        self.print_assets()
        self.print_liabilities()
        self.print_equity()



import os
from decimal import Decimal

from django_ledger.io.ofx import OFXFileManager
from django_ledger.tests.base import DjangoLedgerBaseTest


class SimpleOFXTest(DjangoLedgerBaseTest):
    BASE_PATH = "django_ledger/tests/test_io_ofx/samples/"

    def get_sample_ofx(self, ofx_sample_name: str):
        ofx = OFXFileManager(ofx_file_or_path=os.path.join(self.BASE_PATH, ofx_sample_name))

        return ofx

    def test_ofx_v1_with_intu_bid_field(self):
        """
        OFX v1 with <INTU.BID> field. These are ofx files that are exported for Quickbooks.
        This field can be used to identify the bank in the absence of the <FI.ORG> fields.
        """
        ofx = self.get_sample_ofx("v1_with_intu_bid.ofx")
        accounts = ofx.get_accounts()

        # The bank and fid fields are not provided in this ofx file.
        self.assertIsNone(accounts[0]["fid"])
        self.assertIsNone(accounts[0]["bank"])
        # balance observed from the ofx file
        self.assertEqual(ofx.ofx_data.statements[0].balance.balamt, Decimal("123456.49"))

    def test_ofx_v1_with_open_tags(self):
        """
        OFX v1 with open tags like `<DTSERVER>20211015063225[-5:EST]` instead of `<DTSERVER>20230510120000</DTSERVER>`
        """
        ofx = self.get_sample_ofx("v1_with_open_tags.ofx")
        accounts = ofx.get_accounts()
        account = accounts[0]

        self.assertIsNone(account["fid"])
        self.assertIsNone(account["bank"])
        self.assertEqual(ofx.ofx_data.statements[0].balance.balamt, Decimal("1868.27"))

    def test_ofx_v2_good(self):
        """
        ofx v2 uses XML rather than SGML. This is a good ofx v2 file.
        """
        ofx = self.get_sample_ofx("v2_good.ofx")
        accounts = ofx.get_accounts()
        account = accounts[0]

        self.assertEqual(account["fid"], "123456789")
        self.assertEqual(account["bank"], "BANK NAME")
        self.assertEqual(ofx.ofx_data.statements[0].balance.balamt, Decimal("5000.00"))



from django.urls import path

from django_ledger import views

urlpatterns = [
    path('<slug:entity_slug>/list/',
         views.InventoryListView.as_view(),
         name='inventory-list'),
    path('<slug:entity_slug>/recount/',
         views.InventoryRecountView.as_view(),
         name='inventory-recount'),
]


from django.conf import settings

from django_ledger import __version__
from django_ledger.settings import DJANGO_LEDGER_THEME


def django_ledger_context(request):
    return {
        'DEBUG': settings.DEBUG,
        'VERSION': __version__,
        'DJANGO_LEDGER_THEME': DJANGO_LEDGER_THEME,
    }



# Generated by Django 4.2.2 on 2023-09-22 20:05

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_ledger', '0009_delete_stagedtransactionmodel'),
    ]

    operations = [
        migrations.DeleteModel(
            name='ImportJobModel',
        ),
    ]



from datetime import date
from decimal import Decimal
from random import choice
from urllib.parse import urlparse
from uuid import uuid4

from django.contrib.auth import get_user_model
from django.urls import reverse

from django_ledger.io.io_core import get_localdate
from django_ledger.io.roles import ASSET_CA_CASH, ASSET_CA_PREPAID, LIABILITY_CL_DEFERRED_REVENUE, \
    LIABILITY_CL_ACC_PAYABLE
from django_ledger.models import EntityModel, BillModel, VendorModel
from django_ledger.tests.base import DjangoLedgerBaseTest
from django_ledger.urls.bill import urlpatterns as bill_urls

UserModel = get_user_model()


class BillModelTests(DjangoLedgerBaseTest):

    def setUp(self) -> None:
        self.URL_PATTERNS = {
            p.name: set(p.pattern.converters.keys()) for p in bill_urls
        }

    def test_protected_views(self):
        """
        All Bill Model views must have user authenticated.
        If not, user must be redirected to login page.
        """

        self.logout_client()
        entity_model = self.get_random_entity_model()
        bill_model = choice(entity_model.get_bills())

        for path, kwargs in self.URL_PATTERNS.items():
            url_kwargs = dict()
            url_kwargs['entity_slug'] = entity_model.slug
            if 'bill_pk' in kwargs:
                url_kwargs['bill_pk'] = bill_model.uuid
            if 'year' in kwargs:
                url_kwargs['year'] = bill_model.date_draft.year
            if 'month' in kwargs:
                url_kwargs['month'] = bill_model.date_draft.month
            if 'po_pk' in kwargs:
                url_kwargs['po_pk'] = uuid4()
            if 'ce_pk' in kwargs:
                url_kwargs['ce_pk'] = uuid4()

            url = reverse(f'django_ledger:{path}', kwargs=url_kwargs)
            response = self.CLIENT.get(url, follow=False)
            redirect_url = urlparse(response.url)
            redirect_path = redirect_url.path
            login_path = reverse(viewname='django_ledger:login')

            self.assertEqual(response.status_code, 302, msg=f'{path} view is not protected.')
            self.assertEqual(redirect_path, login_path, msg=f'{path} view not redirecting to correct auth URL.')

    def create_bill(self, amount: Decimal, draft_date: date = None, is_accrued: bool = False) -> tuple[
        EntityModel, BillModel]:
        entity_model: EntityModel = choice(self.ENTITY_MODEL_QUERYSET)
        vendor_model: VendorModel = choice(entity_model.get_vendors())
        account_qs = entity_model.get_default_coa_accounts()

        len(account_qs)  # force evaluation

        cash_account = account_qs.filter(role__in=[ASSET_CA_CASH]).first()
        prepaid_account = account_qs.filter(role__in=[ASSET_CA_PREPAID]).first()
        unearned_account = account_qs.filter(role__in=[LIABILITY_CL_DEFERRED_REVENUE]).first()
        dt = self.get_random_date() if not draft_date else draft_date

        bill_model = BillModel()
        ledger_model, bill_model = bill_model.configure(
            entity_slug=entity_model,
            user_model=self.user_model
        )

        bill_model.amount_due = amount
        bill_model.date_draft = dt
        bill_model.vendor = vendor_model
        bill_model.accrue = is_accrued
        bill_model.xref = 'ABC123xref'
        bill_model.cash_account = cash_account
        bill_model.prepaid_account = prepaid_account
        bill_model.unearned_account = unearned_account
        bill_model.clean()
        bill_model.save()

        return entity_model, bill_model



    def test_bill_list(self):

        self.login_client()
        entity_model = choice(self.ENTITY_MODEL_QUERYSET)
        bill_list_url = reverse('django_ledger:bill-list',
                                kwargs={
                                    'entity_slug': entity_model.slug
                                })

        with self.assertNumQueries(6):
            response = self.CLIENT.get(bill_list_url)

            # bill-list view is rendered...
            self.assertEqual(response.status_code, 200)

        bill_model_qs = response.context['bills']

        for bill_model in bill_model_qs:
            bill_detail_url = reverse('django_ledger:bill-detail',
                                      kwargs={
                                          'entity_slug': entity_model.slug,
                                          'bill_pk': bill_model.uuid
                                      })
            bill_update_url = reverse('django_ledger:bill-update',
                                      kwargs={
                                          'entity_slug': entity_model.slug,
                                          'bill_pk': bill_model.uuid
                                      })
            # bill_delete_url = reverse('django_ledger:bill-delete',
            #                           kwargs={
            #                               'entity_slug': entity_model.slug,
            #                               'bill_pk': bill_model.uuid
            #                           })
            # mark_as_paid_url = reverse('django_ledger:bill-mark-paid',
            #                            kwargs={
            #                                'entity_slug': entity_model.slug,
            #                                'bill_pk': bill_model.uuid
            #                            })

            # bill shows in list...
            self.assertContains(response, bill_model.get_html_id(), status_code=200)

            # amount due shows in list...
            # amt_due_fmt = number_format(bill_model.amount_due, decimal_pos=2, use_l10n=True, force_grouping=True)
            needle = f'id="{bill_model.get_html_amount_due_id()}"'
            self.assertContains(response, needle)

            # amount paid shows in list...
            needle = f'id="{bill_model.get_html_amount_paid_id()}"'
            self.assertContains(response, needle)

            # contains bill-detail url
            self.assertContains(response, bill_detail_url)
            # contains bill-update url
            self.assertContains(response, bill_update_url)
            # contains bill-delete url
            # self.assertContains(response, bill_delete_url)

            # if bill_model.is_approved() and not bill_model.is_paid():
            #     # shows link to mark as paid...
            #     self.assertContains(response, mark_as_paid_url)
            #     with self.assertNumQueries(11):
            #         paid_response = self.CLIENT.get(mark_as_paid_url, follow=False)
            #     self.assertRedirects(paid_response, expected_url=bill_update_url)

            # elif bill_model.is_approved() and bill_model.is_paid():
            #     # if paid, it cannot be paid
            #     self.assertNotContains(response, mark_as_paid_url)

        # # making one payment...
        # bill_model.make_payment(amt=Decimal('250.50'),
        #                         entity_slug=entity_model.slug,
        #                         user_model=self.user_model,
        #                         commit=True)
        #
        # response = self.CLIENT.get(bill_list_url)
        # amt_paid = number_format(bill_model.amount_paid, decimal_pos=2, use_l10n=True, force_grouping=True)
        # needle = f'<td id="{bill_model.get_html_amount_paid_id()}">${amt_paid}</td>'
        # self.assertEqual(bill_model.amount_paid, Decimal('250.50'))
        # self.assertContains(response, needle)
        #
        # # making additional payment...
        # bill_model.make_payment(amt=Decimal('125.60'),
        #                         entity_slug=entity_model.slug,
        #                         user_model=self.user_model,
        #                         commit=True)
        # response = self.CLIENT.get(bill_list_url)
        # amt_paid = number_format(bill_model.amount_paid, decimal_pos=2, use_l10n=True, force_grouping=True)
        # needle = f'<td id="{bill_model.get_html_amount_paid_id()}">${amt_paid}</td>'
        # self.assertEqual(bill_model.amount_paid, Decimal('376.10'))
        # self.assertContains(response, needle)
        #
        # # mark as paid...
        # bill_model.mark_as_paid(entity_slug=entity_model.slug,
        #                         user_model=self.user_model,
        #                         commit=True)
        # response = self.CLIENT.get(bill_list_url)
        # amt_paid = number_format(bill_model.amount_paid, decimal_pos=2, use_l10n=True, force_grouping=True)
        # needle = f'<td id="{bill_model.get_html_amount_paid_id()}">${amt_paid}</td>'
        # self.assertEqual(bill_model.amount_paid, bill_model.amount_due)
        # self.assertContains(response, needle)

    def test_bill_create(self):

        self.login_client()
        entity_model: EntityModel = choice(self.ENTITY_MODEL_QUERYSET)

        bill_create_url = reverse('django_ledger:bill-create',
                                  kwargs={
                                      'entity_slug': entity_model.slug
                                  })

        response = self.CLIENT.get(bill_create_url)

        # bill create form is rendered
        self.assertContains(response,
                            f'id="djl-bill-model-create-form-id"',
                            msg_prefix='Bill create form is not rendered.')

        # user can select a vendor
        self.assertContains(response,
                            'id="djl-bill-create-vendor-select-input"',
                            msg_prefix='Vendor Select input not rendered.')

        # user can select bill terms...
        self.assertContains(response,
                            'id="djl-bill-create-terms-select-input"',
                            msg_prefix='Bill terms input not rendered.')

        # user can select bill terms...
        self.assertContains(response,
                            'id="djl-bill-xref-input"',
                            msg_prefix='Bill XREF input not rendered.')

        # user can select date...
        self.assertContains(response,
                            'id="djl-bill-draft-date-input"',
                            msg_prefix='Bill draft date input not rendered.')

        # user can select cash account...
        self.assertContains(response,
                            'id="djl-bill-cash-account-input"',
                            msg_prefix='Bill cash account input not rendered.')

        # user can select prepaid account...
        self.assertContains(response,
                            'id="djl-bill-prepaid-account-input"',
                            msg_prefix='Bill prepaid account input not rendered.')

        # user can select unearned account...
        self.assertContains(response,
                            'id="djl-bill-unearned-account-input"',
                            msg_prefix='Bill unearned account input not rendered.')

        # user can select unearned account...
        self.assertContains(response,
                            'id="djl-bill-create-button"',
                            msg_prefix='Bill create button not rendered.')

        # user cannot input amount due...
        self.assertNotContains(response,
                               'id="djl-bill-amount-due-input"',
                               msg_prefix='Bill amount due input not rendered.')

        # user can navigate to bill list
        bill_list_url = reverse('django_ledger:bill-list',
                                kwargs={
                                    'entity_slug': entity_model.slug
                                })
        self.assertContains(response, bill_list_url)

        account_qs = entity_model.get_default_coa_accounts()
        len(account_qs)

        a_vendor_model = choice(entity_model.get_vendors())

        bill_data = {
            'vendor': a_vendor_model.uuid,
            'date_draft': get_localdate(),
            'terms': BillModel.TERMS_NET_30,
            'cash_account_id': account_qs.filter(role__exact=ASSET_CA_CASH),
            'prepaid_account_id': account_qs.filter(role__exact=ASSET_CA_PREPAID),
            'unearned_account_id': account_qs.filter(role__exact=LIABILITY_CL_ACC_PAYABLE),
        }

        # create_response = self.CLIENT.post(bill_create_url, data=bill_data, follow=True)
        # self.assert
        # self.assertFormError(create_response, form='form', field=None,
        #                      errors=['Must provide a cash account.'])
        #
        # bill_data['cash_account'] = account_qs.with_roles(roles=ASSET_CA_CASH).first().uuid
        # create_response = self.CLIENT.post(bill_create_url, data=bill_data, follow=True)
        # self.assertFormError(create_response, form='form', field=None,
        #                      errors=['Must provide all accounts Cash, Prepaid, UnEarned.'])
        #
        # cash_account = account_qs.with_roles(roles=ASSET_CA_PREPAID).first()
        # bill_data['prepaid_account'] = cash_account.uuid
        # create_response = self.CLIENT.post(bill_create_url, data=bill_data, follow=True)
        # self.assertFormError(create_response, form='form', field=None,
        #                      errors=['Must provide all accounts Cash, Prepaid, UnEarned.'])
        #
        # del bill_data['prepaid_account']
        # unearned_account = account_qs.with_roles(roles=LIABILITY_CL_DEFERRED_REVENUE).first()
        # bill_data['unearned_account'] = unearned_account.uuid
        # create_response = self.CLIENT.post(bill_create_url, data=bill_data, follow=True)
        # self.assertFormError(create_response, form='form', field=None,
        #                      errors=['Must provide all accounts Cash, Prepaid, UnEarned.'])
        #
        # bill_data['prepaid_account'] = cash_account.uuid
        # create_response = self.CLIENT.post(bill_create_url, data=bill_data, follow=True)
        # self.assertTrue(create_response.resolver_match.view_name, 'django_ledger:bill-detail')

    def test_bill_detail(self):
        self.login_client()
        # today = get_localdate()

        for i in range(5):
            entity_model: EntityModel = self.get_random_entity_model()
            bill_model: BillModel = choice(entity_model.get_bills())
            vendor_model = bill_model.vendor
            bill_detail_url = bill_model.get_absolute_url()

            with self.assertNumQueries(6):
                bill_detail_response = self.CLIENT.get(bill_detail_url)
            self.assertTrue(bill_detail_response.status_code, 200)

            # self.assertTrue(bill_model.is_draft())
            # 'Not Approved' is displayed to the user...
            # self.assertFalse(bill_model.is_approved())

            # bill card is displayed to the user...
            self.assertContains(bill_detail_response, 'id="djl-bill-card-widget"')

            # vendor card is displayed to the user...
            self.assertContains(bill_detail_response, 'id="djl-vendor-card-widget"')

            # if bill is not approved or draft...
            # user can update bill
            # self.assertContains(bill_detail_response, 'id="djl-bill-detail-update-button"')
            # user cannot mark as paid
            # self.assertNotContains(bill_detail_response, 'id="djl-bill-detail-mark-paid-button"')
            # user can delete..
            # self.assertContains(bill_detail_response, 'id="djl-bill-detail-delete-button"')
            # user cannot void...
            # self.assertNotContains(bill_detail_response, 'id="djl-bill-detail-void-button"')

            # user can navigate to bill-list...
            self.assertContains(bill_detail_response,
                                reverse('django_ledger:bill-list',
                                        kwargs={
                                            'entity_slug': entity_model.slug
                                        }))

            # vendor name is shown
            self.assertContains(bill_detail_response, vendor_model.vendor_name)

            # can edit vendor link
            self.assertContains(bill_detail_response,
                                reverse('django_ledger:vendor-update',
                                        kwargs={
                                            'entity_slug': entity_model.slug,
                                            'vendor_pk': vendor_model.uuid
                                        }))

            # link to cash account detail
            self.assertContains(bill_detail_response,
                                reverse('django_ledger:account-detail',
                                        kwargs={
                                            'entity_slug': entity_model.slug,
                                            'account_pk': bill_model.cash_account.uuid
                                        }))

            # amount paid is shown
            self.assertContains(bill_detail_response, 'id="djl-bill-detail-amount-paid"')

            # amount owed is shown
            self.assertContains(bill_detail_response, 'id="djl-bill-detail-amount-owed"')

            if not bill_model.accrue:
                # amount prepaid is not shown
                self.assertNotContains(bill_detail_response, ' id="djl-bill-detail-amount-prepaid"')
                # amount unearned is not shown
                self.assertNotContains(bill_detail_response, ' id="djl-bill-detail-amount-unearned"')

            else:
                # amount prepaid is shown
                self.assertContains(bill_detail_response, ' id="djl-bill-detail-amount-prepaid"')
                # amount unearned is shown
                self.assertContains(bill_detail_response, ' id="djl-bill-detail-amount-unearned"')

            # # amounts are zero...
            # self.assertEqual(bill_model.get_amount_cash(), Decimal('0.00'))
            # self.assertEqual(bill_model.get_amount_earned(), Decimal('0.00'))
            # self.assertEqual(bill_model.get_amount_open(), Decimal('0.00'))
            # self.assertEqual(bill_model.get_amount_prepaid(), Decimal('0.00'))
            # self.assertEqual(bill_model.get_amount_unearned(), Decimal('0.00'))



from django.urls import path

from django_ledger import views

urlpatterns = [
    path('<slug:entity_slug>/latest/',
         views.BillModelListView.as_view(),
         name='bill-list'),
    path('<slug:entity_slug>/year/<int:year>/',
         views.BillModelYearListView.as_view(),
         name='bill-list-year'),
    path('<slug:entity_slug>/month/<int:year>/<int:month>/',
         views.BillModelMonthListView.as_view(),
         name='bill-list-month'),
    path('<slug:entity_slug>/create/',
         views.BillModelCreateView.as_view(),
         name='bill-create'),
    path('<slug:entity_slug>/create/purchase-order/<uuid:po_pk>/',
         views.BillModelCreateView.as_view(for_purchase_order=True),
         name='bill-create-po'),
    path('<slug:entity_slug>/create/estimate/<uuid:ce_pk>/',
         views.BillModelCreateView.as_view(for_estimate=True),
         name='bill-create-estimate'),
    path('<slug:entity_slug>/detail/<uuid:bill_pk>/',
         views.BillModelDetailView.as_view(),
         name='bill-detail'),
    path('<slug:entity_slug>/update/<uuid:bill_pk>/',
         views.BillModelUpdateView.as_view(),
         name='bill-update'),
    path('<slug:entity_slug>/update/<uuid:bill_pk>/items/',
         views.BillModelUpdateView.as_view(action_update_items=True),
         name='bill-update-items'),

    # Actions...
    path('<slug:entity_slug>/actions/<uuid:bill_pk>/mark-as-draft/',
         views.BillModelActionMarkAsDraftView.as_view(),
         name='bill-action-mark-as-draft'),
    path('<slug:entity_slug>/actions/<uuid:bill_pk>/mark-as-review/',
         views.BillModelActionMarkAsInReviewView.as_view(),
         name='bill-action-mark-as-review'),
    path('<slug:entity_slug>/actions/<uuid:bill_pk>/mark-as-approved/',
         views.BillModelActionMarkAsApprovedView.as_view(),
         name='bill-action-mark-as-approved'),
    path('<slug:entity_slug>/actions/<uuid:bill_pk>/mark-as-paid/',
         views.BillModelActionMarkAsPaidView.as_view(),
         name='bill-action-mark-as-paid'),
    path('<slug:entity_slug>/actions/<uuid:bill_pk>/mark-as-void/',
         views.BillModelActionVoidView.as_view(),
         name='bill-action-mark-as-void'),
    path('<slug:entity_slug>/actions/<uuid:bill_pk>/mark-as-canceled/',
         views.BillModelActionCanceledView.as_view(),
         name='bill-action-mark-as-canceled'),
    path('<slug:entity_slug>/actions/<uuid:bill_pk>/lock-ledger/',
         views.BillModelActionLockLedgerView.as_view(),
         name='bill-action-lock-ledger'),
    path('<slug:entity_slug>/actions/<uuid:bill_pk>/unlock-ledger/',
         views.BillModelActionUnlockLedgerView.as_view(),
         name='bill-action-unlock-ledger'),
    path('<slug:entity_slug>/actions/<uuid:bill_pk>/force-migration/',
         views.BillModelActionForceMigrateView.as_view(),
         name='bill-action-force-migrate'),
]



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""
from collections import defaultdict
from itertools import groupby, chain

from django.core.exceptions import ValidationError

from django_ledger.io import roles as roles_module
from django_ledger.models.utils import lazy_loader


class AccountRoleIOMiddleware:

    def __init__(self,
                 io_data: dict,
                 by_period: bool = False,
                 by_unit: bool = False):

        self.BY_PERIOD = by_period
        self.BY_UNIT = by_unit

        self.DIGEST = io_data
        self.DIGEST['role_account'] = None
        self.DIGEST['role_balance'] = None

        self.ACCOUNTS = io_data['accounts']

        self.ROLES_ACCOUNTS = dict()
        self.ROLES_BALANCES = dict()
        self.ROLES_BALANCE_SHEET = dict()

        if self.BY_PERIOD:
            self.ROLES_BALANCES_BY_PERIOD = defaultdict(lambda: dict())
            self.DIGEST['role_balance_by_period'] = None
        if self.BY_UNIT:
            self.ROLES_BALANCES_BY_UNIT = defaultdict(lambda: dict())
            self.DIGEST['role_balance_by_unit'] = None

        if self.BY_PERIOD and self.BY_UNIT:
            self.ROLES_BALANCES_BY_PERIOD_AND_UNIT = defaultdict(lambda: dict())

    def digest(self):

        self.process_roles()
        self.DIGEST['role_account'] = self.ROLES_ACCOUNTS
        self.DIGEST['role_balance'] = self.ROLES_BALANCES

        if self.BY_PERIOD:
            self.DIGEST['role_balance_by_period'] = self.ROLES_BALANCES_BY_PERIOD
        if self.BY_UNIT:
            self.DIGEST['role_balance_by_unit'] = self.ROLES_BALANCES_BY_UNIT

        return self.DIGEST

    def process_roles(self):

        for c, l in roles_module.ROLES_DIRECTORY.items():
            for r in l:
                acc_list = list(acc for acc in self.ACCOUNTS if acc['role'] == getattr(roles_module, r))

                self.ROLES_ACCOUNTS[r] = acc_list
                self.ROLES_BALANCES[r] = sum(acc['balance'] for acc in acc_list)

                if self.BY_PERIOD or self.BY_UNIT:
                    for acc in acc_list:
                        if self.BY_PERIOD:
                            key = (acc['period_year'], acc['period_month'])
                            self.ROLES_BALANCES_BY_PERIOD[key][r] = sum(acc['balance'] for acc in acc_list if all([
                                acc['period_year'] == key[0],
                                acc['period_month'] == key[1]]
                            ))
                        if self.BY_UNIT:
                            key = (acc['unit_uuid'], acc['unit_name'])
                            self.ROLES_BALANCES_BY_UNIT[key][r] = sum(
                                acc['balance'] for acc in acc_list if acc['unit_uuid'] == key[0])


class AccountGroupIOMiddleware:
    GROUP_ACCOUNTS_KEY = 'group_account'
    GROUP_BALANCE_KEY = 'group_balance'
    GROUP_BALANCE_BY_UNIT_KEY = 'group_balance_by_unit'
    GROUP_BALANCE_BY_PERIOD_KEY = 'group_balance_by_period'

    def __init__(self,
                 io_data: dict,
                 by_period: bool = False,
                 by_unit: bool = False):

        self.BY_PERIOD = by_period
        self.BY_UNIT = by_unit

        self.IO_DIGEST = io_data

        self.IO_DIGEST[self.GROUP_ACCOUNTS_KEY] = None
        self.IO_DIGEST[self.GROUP_BALANCE_KEY] = None

        self.DIGEST_ACCOUNTS = io_data['accounts']

        self.GROUPS_ACCOUNTS = dict()
        self.GROUPS_BALANCES = dict()

        if self.BY_PERIOD:
            self.GROUPS_BALANCES_BY_PERIOD = defaultdict(lambda: dict())
            self.IO_DIGEST[self.GROUP_BALANCE_BY_PERIOD_KEY] = None

        if self.BY_UNIT:
            self.GROUPS_BALANCES_BY_UNIT = defaultdict(lambda: dict())
            self.IO_DIGEST[self.GROUP_BALANCE_BY_UNIT_KEY] = None

        if self.BY_PERIOD and self.BY_UNIT:
            self.GROUPS_BALANCES_BY_PERIOD_AND_UNIT = defaultdict(lambda: dict())
            self.IO_DIGEST[self.GROUP_BALANCE_BY_PERIOD_KEY] = None

    def digest(self):

        self.process_groups()
        self.IO_DIGEST[self.GROUP_ACCOUNTS_KEY] = self.GROUPS_ACCOUNTS
        self.IO_DIGEST[self.GROUP_BALANCE_KEY] = self.GROUPS_BALANCES

        if self.BY_PERIOD:
            self.IO_DIGEST[self.GROUP_BALANCE_BY_PERIOD_KEY] = self.GROUPS_BALANCES_BY_PERIOD
        if self.BY_UNIT:
            self.IO_DIGEST[self.GROUP_BALANCE_BY_UNIT_KEY] = self.GROUPS_BALANCES_BY_UNIT
        return self.IO_DIGEST

    def get_accounts_generator(self, mod, g):
        return (acc for acc in self.DIGEST_ACCOUNTS if acc['role'] in getattr(mod, g))

    def process_groups(self):
        for g in roles_module.ROLES_GROUPS:
            acc_list = list(self.get_accounts_generator(roles_module, g))
            self.GROUPS_ACCOUNTS[g] = acc_list
            self.GROUPS_BALANCES[g] = sum(acc['balance'] for acc in acc_list)

            if self.BY_PERIOD or self.BY_UNIT:
                for acc in acc_list:
                    if self.BY_PERIOD:
                        key = (acc['period_year'], acc['period_month'])
                        self.GROUPS_BALANCES_BY_PERIOD[key][g] = sum(
                            acc['balance'] for acc in acc_list if all([
                                acc['period_year'] == key[0],
                                acc['period_month'] == key[1]]
                            ))
                    if self.BY_UNIT:
                        key = (acc['unit_uuid'], acc['unit_name'])
                        self.GROUPS_BALANCES_BY_UNIT[key][g] = sum(
                            acc['balance'] for acc in acc_list if acc['unit_uuid'] == key[0]
                        )


class JEActivityIOMiddleware:

    def __init__(self,
                 io_data: dict,
                 by_unit: bool = False,
                 by_period: bool = False):

        self.DIGEST = io_data
        self.DIGEST['activity_account'] = None
        self.DIGEST['activity_balance'] = None

        self.BY_PERIOD = by_period
        self.BY_UNIT = by_unit

        self.ACCOUNTS = io_data['accounts']
        self.ACTIVITY_ACCOUNTS = dict()
        self.ACTIVITY_BALANCES = dict()

        if self.BY_PERIOD:
            self.ACTIVITY_BALANCES_BY_PERIOD = defaultdict(lambda: dict())
            self.DIGEST['activity_balance_by_period'] = None
        if self.BY_UNIT:
            self.ACTIVITY_BALANCES_BY_UNIT = defaultdict(lambda: dict())
            self.DIGEST['activity_balance_by_unit'] = None
        if self.BY_PERIOD and self.BY_UNIT:
            self.ROLES_BALANCES_BY_PERIOD_AND_UNIT = defaultdict(lambda: dict())

    def digest(self):

        self.process_activity()
        self.DIGEST['activity_account'] = self.ACTIVITY_ACCOUNTS
        self.DIGEST['activity_balance'] = self.ACTIVITY_BALANCES

        if self.BY_PERIOD:
            self.DIGEST['activity_balance_by_period'] = self.ACTIVITY_BALANCES_BY_PERIOD
        if self.BY_UNIT:
            self.DIGEST['activity_balance_by_unit'] = self.ACTIVITY_BALANCES_BY_PERIOD

    def get_accounts_generator(self, activity: str):
        return (acc for acc in self.ACCOUNTS if acc['activity'] == activity)

    def process_activity(self):
        JournalEntryModel = lazy_loader.get_journal_entry_model()
        for act in JournalEntryModel.VALID_ACTIVITIES:
            acc_list = list(self.get_accounts_generator(act))
            self.ACTIVITY_ACCOUNTS[act] = acc_list
            self.ACTIVITY_BALANCES[act] = sum(acc['balance'] for acc in acc_list)

            if self.BY_PERIOD or self.BY_UNIT:
                for acc in acc_list:
                    if self.BY_PERIOD:
                        key = (acc['period_year'], acc['period_month'])
                        self.ACTIVITY_BALANCES_BY_PERIOD[key][act] = sum(acc['balance'] for acc in acc_list if all([
                            acc['period_year'] == key[0],
                            acc['period_month'] == key[1]]
                        ))
                    if self.BY_UNIT:
                        key = (acc['unit_uuid'], acc['unit_name'])
                        self.ACTIVITY_BALANCES_BY_UNIT[key][act] = sum(
                            acc['balance'] for acc in acc_list if acc['unit_uuid'] == key[0])


class BalanceSheetIOMiddleware:
    BS_DIGEST_KEY = 'balance_sheet'

    def __init__(self, io_data: dict):
        self.IO_DATA = io_data

    def digest(self):
        if 'group_account' in self.IO_DATA:
            gb_bs = {
                bsr: list(l) for bsr, l in groupby(
                    chain.from_iterable(
                        [
                            self.IO_DATA['group_account']['GROUP_ASSETS'],
                            self.IO_DATA['group_account']['GROUP_LIABILITIES'],
                            self.IO_DATA['group_account']['GROUP_CAPITAL'],
                        ]
                    ),
                    key=lambda acc: acc['role_bs'])
            }

            bs_context = {
                bs_role: {
                    'total_balance': sum(a['balance'] for a in gb),
                    'is_block': True,
                    'roles': {
                        r: {
                            'accounts': list(a)
                        } for r, a in groupby(list(gb), key=lambda acc: acc['role'])
                    }
                } for bs_role, gb in gb_bs.items()
            }

            for bs_role, bs_role_data in bs_context.items():
                for acc_role, role_data in bs_role_data['roles'].items():
                    role_data['total_balance'] = sum(a['balance'] for a in role_data['accounts'])
                    role_data['role_name'] = roles_module.ACCOUNT_LIST_ROLE_VERBOSE[acc_role]

            bs_context['equity_balance'] = self.IO_DATA['group_balance']['GROUP_EQUITY']
            bs_context['retained_earnings_balance'] = self.IO_DATA['group_balance']['GROUP_EARNINGS']
            bs_context['liabilities_equity_balance'] = self.IO_DATA['group_balance']['GROUP_LIABILITIES_EQUITY']

            self.IO_DATA[self.BS_DIGEST_KEY] = bs_context

        return self.IO_DATA


class IncomeStatementIOMiddleware:
    IC_DIGEST_KEY = 'income_statement'

    def __init__(self, io_data: dict):
        self.IO_DATA = io_data

    def digest(self):
        if 'group_account' in self.IO_DATA:
            self.IO_DATA[self.IC_DIGEST_KEY] = {
                'operating': {
                    'revenues': [
                        acc for acc in self.IO_DATA['group_account']['GROUP_INCOME'] if
                        acc['role'] in roles_module.GROUP_IC_OPERATING_REVENUES
                    ],
                    'cogs': [
                        acc for acc in self.IO_DATA['group_account']['GROUP_COGS'] if
                        acc['role'] in roles_module.GROUP_IC_OPERATING_COGS
                    ],
                    'expenses': [
                        acc for acc in self.IO_DATA['group_account']['GROUP_EXPENSES'] if
                        acc['role'] in roles_module.GROUP_IC_OPERATING_EXPENSES
                    ]
                },
                'other': {
                    'revenues': [acc for acc in self.IO_DATA['group_account']['GROUP_INCOME'] if
                                 acc['role'] in roles_module.GROUP_IC_OTHER_REVENUES],
                    'expenses': [acc for acc in self.IO_DATA['group_account']['GROUP_EXPENSES'] if
                                 acc['role'] in roles_module.GROUP_IC_OTHER_EXPENSES],
                }
            }

            for activity, ic_section in self.IO_DATA[self.IC_DIGEST_KEY].items():
                for section, acc_list in ic_section.items():
                    for acc in acc_list:
                        acc['role_name'] = roles_module.ACCOUNT_LIST_ROLE_VERBOSE[acc['role']]

            # OPERATING INCOME...
            self.IO_DATA[self.IC_DIGEST_KEY]['operating']['gross_profit'] = sum(
                acc['balance'] for acc in chain.from_iterable(
                    [
                        self.IO_DATA[self.IC_DIGEST_KEY]['operating']['revenues'],
                        self.IO_DATA[self.IC_DIGEST_KEY]['operating']['cogs']
                    ]
                ))
            self.IO_DATA[self.IC_DIGEST_KEY]['operating']['net_operating_income'] = sum(
                acc['balance'] for acc in chain.from_iterable(
                    [
                        self.IO_DATA[self.IC_DIGEST_KEY]['operating']['revenues'],
                        self.IO_DATA[self.IC_DIGEST_KEY]['operating']['cogs'],
                        self.IO_DATA[self.IC_DIGEST_KEY]['operating']['expenses'],
                    ]
                ))
            self.IO_DATA[self.IC_DIGEST_KEY]['operating']['net_operating_revenue'] = sum(
                acc['balance'] for acc in self.IO_DATA[self.IC_DIGEST_KEY]['operating']['revenues']
            )
            self.IO_DATA[self.IC_DIGEST_KEY]['operating']['net_cogs'] = sum(
                acc['balance'] for acc in self.IO_DATA[self.IC_DIGEST_KEY]['operating']['cogs']
            )
            self.IO_DATA[self.IC_DIGEST_KEY]['operating']['net_operating_expenses'] = sum(
                acc['balance'] for acc in self.IO_DATA[self.IC_DIGEST_KEY]['operating']['expenses']
            )

            # OTHER INCOME....
            self.IO_DATA[self.IC_DIGEST_KEY]['other']['net_other_revenues'] = sum(
                acc['balance'] for acc in self.IO_DATA[self.IC_DIGEST_KEY]['other']['revenues']
            )
            self.IO_DATA[self.IC_DIGEST_KEY]['other']['net_other_expenses'] = sum(
                acc['balance'] for acc in self.IO_DATA[self.IC_DIGEST_KEY]['other']['expenses']
            )
            self.IO_DATA[self.IC_DIGEST_KEY]['other']['net_other_income'] = sum(
                acc['balance'] for acc in chain.from_iterable(
                    [
                        self.IO_DATA[self.IC_DIGEST_KEY]['other']['revenues'],
                        self.IO_DATA[self.IC_DIGEST_KEY]['other']['expenses']
                    ]
                ))

            # NET INCOME...
            self.IO_DATA[self.IC_DIGEST_KEY]['net_income'] = self.IO_DATA[self.IC_DIGEST_KEY]['operating'][
                'net_operating_income']
            self.IO_DATA[self.IC_DIGEST_KEY]['net_income'] += self.IO_DATA[self.IC_DIGEST_KEY]['other'][
                'net_other_income']
        return self.IO_DATA


class CashFlowStatementIOMiddleware:
    CFS_DIGEST_KEY = 'cash_flow_statement'

    def __init__(self, io_data: dict):
        self.IO_DATA = io_data
        self.CASH_ACCOUNTS = [a for a in self.IO_DATA['accounts'] if a['role'] == roles_module.ASSET_CA_CASH]
        self.JE_MODEL = lazy_loader.get_journal_entry_model()

    def check_io_digest(self):
        if AccountGroupIOMiddleware.GROUP_BALANCE_KEY not in self.IO_DATA:
            raise ValidationError(
                'IO Digest must have groups for Cash Flow Statement'
            )

    def operating(self):
        group_balances = self.IO_DATA[AccountGroupIOMiddleware.GROUP_BALANCE_KEY]
        operating_activities = dict()
        operating_activities['GROUP_CFS_NET_INCOME'] = {
            'description': 'Net Income',
            'balance': group_balances['GROUP_CFS_NET_INCOME']
        }
        operating_activities['GROUP_CFS_OP_DEPRECIATION_AMORTIZATION'] = {
            'description': 'Depreciation & Amortization of Assets',
            'balance': -group_balances['GROUP_CFS_OP_DEPRECIATION_AMORTIZATION']
        }
        operating_activities['GROUP_CFS_OP_INVESTMENT_GAINS'] = {
            'description': 'Gain/Loss Sale of Assets',
            'balance': group_balances['GROUP_CFS_OP_INVESTMENT_GAINS']
        }
        operating_activities['GROUP_CFS_OP_ACCOUNTS_RECEIVABLE'] = {
            'description': 'Accounts Receivable',
            'balance': -group_balances['GROUP_CFS_OP_ACCOUNTS_RECEIVABLE']
        }
        operating_activities['GROUP_CFS_OP_INVENTORY'] = {
            'description': 'Inventories',
            'balance': -group_balances['GROUP_CFS_OP_INVENTORY']
        }

        operating_activities['GROUP_CFS_OP_ACCOUNTS_PAYABLE'] = {
            'description': 'Accounts Payable',
            'balance': group_balances['GROUP_CFS_OP_ACCOUNTS_PAYABLE']
        }
        operating_activities['GROUP_CFS_OP_OTHER_CURRENT_ASSETS_ADJUSTMENT'] = {
            'description': 'Other Current Assets',
            'balance': -group_balances['GROUP_CFS_OP_OTHER_CURRENT_ASSETS_ADJUSTMENT']
        }
        operating_activities['GROUP_CFS_OP_OTHER_CURRENT_LIABILITIES_ADJUSTMENT'] = {
            'description': 'Other Current Liabilities',
            'balance': group_balances['GROUP_CFS_OP_OTHER_CURRENT_LIABILITIES_ADJUSTMENT']
        }

        net_cash_by_op_activities = sum(i['balance'] for g, i in operating_activities.items())
        self.IO_DATA[self.CFS_DIGEST_KEY]['operating'] = operating_activities
        self.IO_DATA[self.CFS_DIGEST_KEY]['net_cash_by_activity'] = dict(
            OPERATING=net_cash_by_op_activities
        )

    def financing(self):
        # group_balances = self.IO_DIGEST[GroupContextManager.GROUP_BALANCE_KEY]
        financing_activities = dict()
        financing_activities['GROUP_CFS_FIN_ISSUING_EQUITY'] = {
            'description': 'Common Stock, Preferred Stock and Capital Raised',
            'balance': sum(a['balance'] for a in self.CASH_ACCOUNTS if a['activity'] == self.JE_MODEL.FINANCING_EQUITY)
        }
        financing_activities['GROUP_CFS_FIN_DIVIDENDS'] = {
            'description': 'Dividends Payed Out to Shareholders',
            'balance': sum(
                a['balance'] for a in self.CASH_ACCOUNTS if a['activity'] == self.JE_MODEL.FINANCING_DIVIDENDS)
        }
        financing_activities['GROUP_CFS_FIN_ST_DEBT_PAYMENTS'] = {
            'description': 'Increase/Reduction of Short-Term Debt Principal',
            'balance': sum(a['balance'] for a in self.CASH_ACCOUNTS if a['activity'] == self.JE_MODEL.FINANCING_STD)
        }
        financing_activities['GROUP_CFS_FIN_LT_DEBT_PAYMENTS'] = {
            'description': 'Increase/Reduction of Long-Term Debt Principal',
            'balance': sum(a['balance'] for a in self.CASH_ACCOUNTS if a['activity'] == self.JE_MODEL.FINANCING_LTD)
        }

        net_cash = sum(i['balance'] for g, i in financing_activities.items())
        self.IO_DATA[self.CFS_DIGEST_KEY]['financing'] = financing_activities
        self.IO_DATA[self.CFS_DIGEST_KEY]['net_cash_by_activity']['FINANCING'] = net_cash

    def investing(self):
        group_balances = self.IO_DATA[AccountGroupIOMiddleware.GROUP_BALANCE_KEY]
        investing_activities = dict()
        investing_activities['GROUP_CFS_INVESTING_SECURITIES'] = {
            'description': 'Purchase, Maturity and Sales of Investments & Securities',
            'balance': sum(
                a['balance'] for a in self.CASH_ACCOUNTS if a['activity'] == self.JE_MODEL.INVESTING_SECURITIES)
        }
        investing_activities['GROUP_CFS_INVESTING_PPE'] = {
            'description': 'Addition and Disposition of Property, Plant & Equipment',
            'balance': sum(
                a['balance'] for a in self.CASH_ACCOUNTS if a['activity'] == self.JE_MODEL.INVESTING_PPE)
        }

        net_cash = sum(i['balance'] for g, i in investing_activities.items())
        self.IO_DATA[self.CFS_DIGEST_KEY]['investing'] = investing_activities
        self.IO_DATA[self.CFS_DIGEST_KEY]['net_cash_by_activity']['INVESTING'] = net_cash

    def net_cash(self):
        self.IO_DATA[self.CFS_DIGEST_KEY]['net_cash'] = sum([
            bal for act, bal in self.IO_DATA[self.CFS_DIGEST_KEY]['net_cash_by_activity'].items()
        ])

    def net_income(self):
        group_balances = self.IO_DATA[AccountGroupIOMiddleware.GROUP_BALANCE_KEY]
        self.IO_DATA[self.CFS_DIGEST_KEY]['net_income'] = group_balances['GROUP_CFS_NET_INCOME']

    def digest(self):
        self.IO_DATA[self.CFS_DIGEST_KEY] = dict()
        self.check_io_digest()
        self.operating()
        self.financing()
        self.investing()
        self.net_income()
        self.net_cash()
        return self.IO_DATA



from django.forms import (ModelForm, DateInput, TextInput, Select,
                          CheckboxInput, BaseModelFormSet,
                          modelformset_factory, Textarea, DateTimeInput)
from django.forms import ValidationError
from django.utils.translation import gettext_lazy as _

from django_ledger.io.roles import ASSET_CA_CASH, ASSET_CA_PREPAID, LIABILITY_CL_ACC_PAYABLE
from django_ledger.models import (ItemModel, AccountModel, BillModel, ItemTransactionModel,
                                  VendorModel, EntityUnitModel, EntityModel)
from django_ledger.settings import DJANGO_LEDGER_FORM_INPUT_CLASSES


class BillModelCreateForm(ModelForm):
    def __init__(self, *args, entity_model: EntityModel, **kwargs):
        super().__init__(*args, **kwargs)
        self.ENTITY_MODEL = entity_model
        self.get_vendor_queryset()
        self.get_accounts_queryset()

    def get_vendor_queryset(self):
        if 'vendor' in self.fields:
            vendor_qs = self.ENTITY_MODEL.vendormodel_set.active()
            self.fields['vendor'].queryset = vendor_qs

    def get_accounts_queryset(self):

        if all([
            'cash_account' in self.fields,
            'prepaid_account' in self.fields,
            'unearned_account' in self.fields,
        ]):
            account_qs = self.ENTITY_MODEL.default_coa.accountmodel_set.all().for_bill()
            self.fields['cash_account'].queryset = account_qs.filter(role__exact=ASSET_CA_CASH)
            self.fields['prepaid_account'].queryset = account_qs.filter(role__exact=ASSET_CA_PREPAID)
            self.fields['unearned_account'].queryset = account_qs.filter(role__exact=LIABILITY_CL_ACC_PAYABLE)

    class Meta:
        model = BillModel
        fields = [
            'vendor',
            'xref',
            'date_draft',
            'terms',
            'cash_account',
            'prepaid_account',
            'unearned_account',
        ]
        labels = {
            'date_draft': _('Draft Date'),
            'unearned_account': _('Payable Account'),
            'prepaid_account': _('Prepaid Expenses Account')
        }
        widgets = {
            'date_draft': DateInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': _('Bill Date (YYYY-MM-DD)...'),
                'id': 'djl-bill-draft-date-input'
            }),
            'amount_due': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': '$$$',
                'id': 'djl-bill-amount-due-input'}),
            'xref': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-large',
                'placeholder': 'External Reference Number...',
                'id': 'djl-bill-xref-input'
            }),
            'terms': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'id': 'djl-bill-create-terms-select-input'
            }),
            'vendor': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'id': 'djl-bill-create-vendor-select-input'
            }),

            'cash_account': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'id': 'djl-bill-cash-account-input'
            }),
            'prepaid_account': Select(
                attrs={
                    'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                    'id': 'djl-bill-prepaid-account-input'
                }),
            'unearned_account': Select(
                attrs={
                    'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                    'id': 'djl-bill-unearned-account-input'
                }),
        }


class BaseBillModelUpdateForm(BillModelCreateForm):

    def __init__(self,
                 *args,
                 entity_model,
                 user_model,
                 **kwargs):
        super().__init__(entity_model=entity_model, *args, **kwargs)
        self.ENTITY_MODEL = entity_model
        self.USER_MODEL = user_model
        self.BILL_MODEL: BillModel = self.instance

    def save(self, commit=True):
        if commit:
            self.BILL_MODEL.update_state()
            self.instance.migrate_state(
                user_model=self.USER_MODEL,
                entity_slug=self.ENTITY_MODEL.slug,
                raise_exception=False
            )
        super().save(commit=commit)

    class Meta:
        model = BillModel
        fields = [
            'markdown_notes'
        ]
        widgets = {
            'xref': TextInput(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                                     'placeholder': 'External Reference...'}),
            'timestamp': DateTimeInput(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'amount_due': TextInput(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES, 'placeholder': '$$$'}),
            'terms': Select(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'bill_status': Select(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'date_paid': DateInput(
                attrs={
                    'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                    'placeholder': _('Date (YYYY-MM-DD)...')}
            ),
            'amount_paid': TextInput(
                attrs={
                    'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                }),
            'progress': TextInput(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'accrue': CheckboxInput(attrs={'type': 'checkbox'}),
            'paid': CheckboxInput(attrs={'type': 'checkbox'}),
            'cash_account': Select(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-danger'}),
            'prepaid_account': Select(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-danger'}),
            'unearned_account': Select(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-danger'}),
            'markdown_notes': Textarea(attrs={
                'class': 'textarea'
            }),
            'vendor': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'id': 'djl-bill-create-vendor-select-input'
            }),
        }
        labels = {
            'progress': _('Bill Progress Amount (%)'),
            'accrue': _('Will this Bill be Accrued?'),
            'markdown_notes': _('Notes')
        }


class DraftBillModelUpdateForm(BaseBillModelUpdateForm):
    class Meta(BaseBillModelUpdateForm.Meta):
        fields = [
            'vendor',
            'terms',
            'xref',
            'markdown_notes'
        ]


class InReviewBillModelUpdateForm(BaseBillModelUpdateForm):
    class Meta(BaseBillModelUpdateForm.Meta):
        fields = [
            'xref',
            'markdown_notes'
        ]


class ApprovedBillModelUpdateForm(BaseBillModelUpdateForm):
    class Meta(BaseBillModelUpdateForm.Meta):
        fields = [
            'amount_paid',
            'markdown_notes'
        ]


class AccruedAndApprovedBillModelUpdateForm(BaseBillModelUpdateForm):
    class Meta(BaseBillModelUpdateForm.Meta):
        fields = [
            'progress',
            'amount_paid',
            'markdown_notes'
        ]


class PaidBillModelUpdateForm(BaseBillModelUpdateForm):
    class Meta(BaseBillModelUpdateForm.Meta):
        fields = [
            'markdown_notes'
        ]


class BillModelConfigureForm(BaseBillModelUpdateForm):
    class Meta(BaseBillModelUpdateForm.Meta):
        fields = [
            'xref',
            'amount_due',
            'amount_paid',
            'date_paid',
            'progress',
            'accrue',
            'cash_account',
            'prepaid_account',
            'unearned_account',
        ]


class BillItemTransactionForm(ModelForm):

    def clean(self):
        cleaned_data = super(BillItemTransactionForm, self).clean()
        itemtxs_model: ItemTransactionModel = self.instance
        if itemtxs_model.po_model is not None:
            quantity = cleaned_data['quantity']
            if quantity > itemtxs_model.po_quantity:
                raise ValidationError(f'Cannot bill more than {itemtxs_model.po_quantity} authorized.')
        return cleaned_data

    class Meta:
        model = ItemTransactionModel
        fields = [
            'item_model',
            'unit_cost',
            'entity_unit',
            'quantity',
        ]
        widgets = {
            'item_model': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            }),
            'entity_unit': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            }),
            'unit_cost': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            }),
            'quantity': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            })
        }


class BaseBillItemTransactionFormset(BaseModelFormSet):

    def __init__(self, *args,
                 entity_model: EntityModel,
                 bill_model: BillModel,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.BILL_MODEL = bill_model
        self.ENTITY_MODEL = entity_model
        self.queryset = self.BILL_MODEL.itemtransactionmodel_set.select_related(
            'item_model',
            'po_model',
            'bill_model'
        ).order_by('-total_amount')

        self.items_qs = self.ENTITY_MODEL.itemmodel_set.bills()
        self.entity_unit_qs = self.ENTITY_MODEL.entityunitmodel_set.all()

        for form in self.forms:
            form.fields['item_model'].queryset = self.items_qs
            form.fields['entity_unit'].queryset = self.entity_unit_qs

            if not self.BILL_MODEL.can_edit_items():
                form.fields['item_model'].disabled = True
                form.fields['quantity'].disabled = True
                form.fields['unit_cost'].disabled = True
                form.fields['entity_unit'].disabled = True

            instance: ItemTransactionModel = form.instance
            if instance.po_model_id:
                form.fields['item_model'].disabled = True
                form.fields['entity_unit'].disabled = True


def get_bill_itemtxs_formset_class(bill_model: BillModel):
    BillItemTransactionFormset = modelformset_factory(
        model=ItemTransactionModel,
        form=BillItemTransactionForm,
        formset=BaseBillItemTransactionFormset,
        can_delete=True,
        extra=5
    )
    return BillItemTransactionFormset



from django.contrib.admin import TabularInline, ModelAdmin
from django.db.models import Count
from django.forms import ModelForm, BooleanField, BaseInlineFormSet

from django_ledger.models.accounts import AccountModel
from django_ledger.models.chart_of_accounts import ChartOfAccountModel
from django_ledger.models.entity import EntityModel


class AccountModelInLineForm(ModelForm):
    role_default = BooleanField(initial=False, required=False)

    class Meta:
        model = AccountModel
        fields = []


class AccountModelInLineFormSet(BaseInlineFormSet):

    def save_new(self, form, commit=True):
        setattr(form.instance, self.fk.name, self.instance)
        if commit:
            account_model = AccountModel.add_root(
                instance=super().save_new(form, commit=False)
            )
            return account_model
        return super().save_new(form, commit=False)


class AccountModelInLine(TabularInline):
    extra = 0
    form = AccountModelInLineForm
    formset = AccountModelInLineFormSet
    show_change_link = True
    exclude = [
        'path',
        'depth',
        'numchild'
    ]
    model = AccountModel
    fieldsets = [
        ('', {
            'fields': [
                'role',
                'balance_type',
                'code',
                'name',
                'role_default',
                'locked',
                'active'
            ]
        })
    ]

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.not_coa_root()


class ChartOfAccountsAdminForm(ModelForm):
    assign_as_default = BooleanField(initial=False, required=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance.is_default():
            self.fields['assign_as_default'].initial = True
            self.fields['assign_as_default'].disabled = True

    def save(self, commit=True):
        if commit:
            if self.cleaned_data['assign_as_default']:
                entity_model: EntityModel = self.instance.entity
                entity_model.default_coa = self.instance
                entity_model.save(update_fields=[
                    'default_coa'
                ])
        return super().save(commit=commit)


class ChartOfAccountsInLine(TabularInline):
    form = ChartOfAccountsAdminForm
    model = ChartOfAccountModel
    extra = 0
    show_change_link = True
    fields = [
        'name',
        'active',
        'assign_as_default'
    ]


class ChartOfAccountsModelAdmin(ModelAdmin):
    list_filter = [
        'entity__name',
        'active'
    ]
    list_display = [
        'entity_name',
        'name',
        'slug',
        'active',
        'account_model_count'
    ]
    search_fields = [
        'slug',
        'entity__name'
    ]
    list_display_links = ['name']
    fields = [
        'name',
        'description',
    ]
    inlines = [
        AccountModelInLine
    ]

    class Meta:
        model = ChartOfAccountModel

    def entity_name(self, obj):
        return obj.entity.name

    def get_queryset(self, request):
        qs = ChartOfAccountModel.objects.for_user(user_model=request.user)
        ordering = self.get_ordering(request)
        if ordering:
            qs = qs.order_by(*ordering)
        qs = qs.select_related('entity').annotate(Count('accountmodel'))
        return qs

    def account_model_count(self, obj):
        return obj.accountmodel__count

    account_model_count.short_description = 'Accounts'



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""

from django_ledger.views.account import *
from django_ledger.views.auth import *
from django_ledger.views.bank_account import *
from django_ledger.views.bill import *
from django_ledger.views.chart_of_accounts import *
from django_ledger.views.customer import *
from django_ledger.views.data_import import *
from django_ledger.views.djl_api import *
from django_ledger.views.entity import *
from django_ledger.views.estimate import *
from django_ledger.views.financial_statement import *
from django_ledger.views.home import *
from django_ledger.views.inventory import *
from django_ledger.views.invoice import *
from django_ledger.views.item import *
from django_ledger.views.journal_entry import *
from django_ledger.views.ledger import *
from django_ledger.views.purchase_order import *
from django_ledger.views.transactions import *
from django_ledger.views.unit import *
from django_ledger.views.vendor import *
from django_ledger.views.closing_entry import *



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""

from calendar import month_name

from django.http import JsonResponse
from django.views.generic import View

from django_ledger.models import BillModel, EntityModel, InvoiceModel
from django_ledger.utils import accruable_net_summary
from django_ledger.views.mixins import DjangoLedgerSecurityMixIn, EntityUnitMixIn


# from jsonschema import validate, ValidationError


class PnLAPIView(DjangoLedgerSecurityMixIn, EntityUnitMixIn, View):
    http_method_names = ['get']

    def get(self, request, *args, **kwargs):
        if request.user.is_authenticated:
            entity = EntityModel.objects.for_user(
                user_model=self.request.user).get(
                slug__exact=self.kwargs['entity_slug'])

            unit_slug = self.get_unit_slug()

            io_digest = entity.digest(
                user_model=self.request.user,
                unit_slug=unit_slug,
                equity_only=True,
                signs=False,
                by_period=True,
                process_groups=True,
                from_date=self.request.GET.get('fromDate'),
                to_date=self.request.GET.get('toDate'),

                # todo: For PnL to display proper period values must not use closing entries.
                use_closing_entries=False
            )

            io_data = io_digest.get_io_data()
            group_balance_by_period = io_data['group_balance_by_period']
            group_balance_by_period = dict(sorted((k, v) for k, v in group_balance_by_period.items()))

            entity_data = {
                f'{month_name[k[1]]} {k[0]}': {d: float(f) for d, f in v.items()} for k, v in
                group_balance_by_period.items()}

            entity_pnl = {
                'entity_slug': entity.slug,
                'entity_name': entity.name,
                'pnl_data': entity_data
            }

            return JsonResponse({
                'results': entity_pnl
            })

        return JsonResponse({
            'message': 'Unauthorized'
        }, status=401)


class PayableNetAPIView(DjangoLedgerSecurityMixIn, EntityUnitMixIn, View):
    http_method_names = ['get']

    def get(self, request, *args, **kwargs):
        if request.user.is_authenticated:
            bill_qs = BillModel.objects.for_entity(
                entity_model=self.AUTHORIZED_ENTITY_MODEL
            ).unpaid()

            # todo: implement this...
            # unit_slug = self.get_unit_slug()
            # if unit_slug:
            #     bill_qs.filter(ledger__journal_entry__entity_unit__slug__exact=unit_slug)

            net_summary = accruable_net_summary(bill_qs)
            entity_model = bill_qs.first().ledger.entity
            net_payables = {
                'entity_slug': self.kwargs['entity_slug'],
                'entity_name': entity_model.name,
                'net_payable_data': net_summary
            }

            return JsonResponse({
                'results': net_payables
            })

        return JsonResponse({
            'message': 'Unauthorized'
        }, status=401)


class ReceivableNetAPIView(DjangoLedgerSecurityMixIn, EntityUnitMixIn, View):
    http_method_names = ['get']

    def get(self, request, *args, **kwargs):
        if request.user.is_authenticated:
            invoice_qs = InvoiceModel.objects.for_entity(
                entity_model=self.kwargs['entity_slug']
            ).for_user(self.request.user).unpaid()

            # todo: implement this...
            # unit_slug = self.get_unit_slug()
            # if unit_slug:
            #     invoice_qs.filter(ledger__journal_entry__entity_unit__slug__exact=unit_slug)

            net_summary = accruable_net_summary(invoice_qs)
            entity_model = invoice_qs.first().ledger.entity
            net_receivable = {
                'entity_slug': self.kwargs['entity_slug'],
                'entity_name': entity_model.name,
                'net_receivable_data': net_summary
            }

            return JsonResponse({
                'results': net_receivable
            })

        return JsonResponse({
            'message': 'Unauthorized'
        }, status=401)



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.
"""

from django.apps import apps


class LazyLoader:
    """
    A class that provides lazy-loading functionality for various models in the django_ledger app.

    Attributes:
        app_config (AppConfig): The AppConfig object for the django_ledger app.

        ENTITY_MODEL (str): The name of the entity model.
        LEDGER_MODEL (str): The name of the ledger model.
        JE_MODEL (str): The name of the journal entry model.
        TRANSACTION_MODEL (str): The name of the transaction model.
        ACCOUNT_MODEL (str): The name of the account model.
        COA_MODEL (str): The name of the chart of account model.

        ENTITY_STATE_MODEL (str): The name of the entity state model.
        ENTITY_UNIT_MODEL (str): The name of the entity unit model.
        CLOSING_ENTRY_MODEL (str): The name of the closing entry model.
        CLOSING_ENTRY_TRANSACTION_MODEL (str): The name of the closing entry transaction model.

        BANK_ACCOUNT_MODEL (str): The name of the bank account model.
        PURCHASE_ORDER_MODEL (str): The name of the purchase order model.

        CUSTOMER_MODEL (str): The name of the customer model.
        INVOICE_MODEL (str): The name of the invoice model.
        BILL_MODEL (str): The name of the bill model.
        UOM_MODEL (str): The name of the unit of measure model.
        VENDOR_MODEL (str): The name of the vendor model.
        ESTIMATE_MODEL (str): The name of the estimate model.
        ITEM_MODEL (str): The name of the item model.
        ITEM_TRANSACTION_MODEL (str): The name of the item transaction model.

        ENTITY_DATA_GENERATOR (EntityDataGenerator): The EntityDataGenerator class used for generating entity data.
        BALANCE_SHEET_REPORT_CLASS (BalanceSheetReport): The BalanceSheetReport class used for generating balance sheet reports.
        INCOME_STATEMENT_REPORT_CLASS (IncomeStatementReport): The IncomeStatementReport class used for generating income statement reports.
        CASH_FLOW_STATEMENT_REPORT_CLASS (CashFlowStatementReport): The CashFlowStatementReport class used for generating cash flow statement reports.

    Methods:
        get_entity_model() -> Model: Returns the entity model.
        get_entity_unit_model() -> Model: Returns the entity unit model.
        get_entity_state_model() -> Model: Returns the entity state model.
        get_bank_account_model() -> Model: Returns the bank account model.
        get_account_model() -> Model: Returns the account model.
        get_coa_model() -> Model: Returns the chart of account model.
        get_txs_model() -> Model: Returns the transaction model.
        get_purchase_order_model() -> Model: Returns the purchase order model.
        get_ledger_model() -> Model: Returns the ledger model.
        get_journal_entry_model() -> Model: Returns the journal entry model.
        get_item_model() -> Model: Returns the item model.
        get_item_transaction_model() -> Model: Returns the item transaction model.
        get_customer_model() -> Model: Returns the customer model.
        get_bill_model() -> Model: Returns the bill model.
        get_invoice_model() -> Model: Returns the invoice model.
        get_uom_model() -> Model: Returns the unit of measure model.
        get_vendor_model() -> Model: Returns the vendor model.
        get_estimate_model() -> Model: Returns the estimate model.
        get_closing_entry_model() -> Model: Returns the closing entry model.
        get_closing_entry_transaction_model() -> Model: Returns the closing entry transaction model.
        get_entity_data_generator() -> EntityDataGenerator: Returns the EntityDataGenerator class.
        get_balance_sheet_report_class() -> BalanceSheetReport: Returns the BalanceSheetReport class.
        get_income_statement_report_class() -> IncomeStatementReport: Returns the IncomeStatementReport class.
        get_cash_flow_statement_report_class() -> CashFlowStatementReport: Returns the CashFlowStatementReport class.
    """

    app_config = apps.get_app_config(app_label='django_ledger')

    ENTITY_MODEL = 'entitymodel'
    LEDGER_MODEL = 'ledgermodel'
    JE_MODEL = 'journalentrymodel'
    TRANSACTION_MODEL = 'transactionmodel'
    ACCOUNT_MODEL = 'accountmodel'
    COA_MODEL = 'chartofaccountmodel'

    ENTITY_STATE_MODEL = 'entitystatemodel'
    ENTITY_UNIT_MODEL = 'entityunitmodel'
    CLOSING_ENTRY_MODEL = 'closingentrymodel'
    CLOSING_ENTRY_TRANSACTION_MODEL = 'closingentrytransactionmodel'

    BANK_ACCOUNT_MODEL = 'bankaccountmodel'
    PURCHASE_ORDER_MODEL = 'purchaseordermodel'

    CUSTOMER_MODEL = 'customermodel'
    INVOICE_MODEL = 'invoicemodel'
    BILL_MODEL = 'billmodel'
    UOM_MODEL = 'unitofmeasuremodel'
    VENDOR_MODEL = 'vendormodel'
    ESTIMATE_MODEL = 'estimatemodel'
    ITEM_MODEL = 'itemmodel'
    ITEM_TRANSACTION_MODEL = 'itemtransactionmodel'

    ENTITY_DATA_GENERATOR = None
    BALANCE_SHEET_REPORT_CLASS = None
    INCOME_STATEMENT_REPORT_CLASS = None
    CASH_FLOW_STATEMENT_REPORT_CLASS = None



    def get_entity_model(self):
        return self.app_config.get_model(self.ENTITY_MODEL)

    def get_entity_unit_model(self):
        return self.app_config.get_model(self.ENTITY_UNIT_MODEL)

    def get_entity_state_model(self):
        return self.app_config.get_model(self.ENTITY_STATE_MODEL)

    def get_bank_account_model(self):
        return self.app_config.get_model(self.BANK_ACCOUNT_MODEL)

    def get_account_model(self):
        return self.app_config.get_model(self.ACCOUNT_MODEL)

    def get_coa_model(self):
        return self.app_config.get_model(self.COA_MODEL)

    def get_txs_model(self):
        return self.app_config.get_model(self.TRANSACTION_MODEL)

    def get_purchase_order_model(self):
        return self.app_config.get_model(self.PURCHASE_ORDER_MODEL)

    def get_ledger_model(self):
        self.get_entity_unit_model()
        return self.app_config.get_model(self.LEDGER_MODEL)

    def get_journal_entry_model(self):
        return self.app_config.get_model(self.JE_MODEL)

    def get_item_model(self):
        return self.app_config.get_model(self.ITEM_MODEL)

    def get_item_transaction_model(self):
        return self.app_config.get_model(self.ITEM_TRANSACTION_MODEL)

    def get_customer_model(self):
        return self.app_config.get_model(self.CUSTOMER_MODEL)

    def get_bill_model(self):
        return self.app_config.get_model(self.BILL_MODEL)

    def get_invoice_model(self):
        return self.app_config.get_model(self.INVOICE_MODEL)

    def get_uom_model(self):
        return self.app_config.get_model(self.UOM_MODEL)

    def get_vendor_model(self):
        return self.app_config.get_model(self.VENDOR_MODEL)

    def get_estimate_model(self):
        return self.app_config.get_model(self.ESTIMATE_MODEL)

    def get_closing_entry_model(self):
        return self.app_config.get_model(self.CLOSING_ENTRY_MODEL)

    def get_closing_entry_transaction_model(self):
        return self.app_config.get_model(self.CLOSING_ENTRY_TRANSACTION_MODEL)

    def get_entity_data_generator(self):
        if not self.ENTITY_DATA_GENERATOR:
            from django_ledger.io.io_generator import EntityDataGenerator
            self.ENTITY_DATA_GENERATOR = EntityDataGenerator
        return self.ENTITY_DATA_GENERATOR

    def get_balance_sheet_report_class(self):
        if not self.BALANCE_SHEET_REPORT_CLASS:
            from django_ledger.report.balance_sheet import BalanceSheetReport
            self.BALANCE_SHEET_REPORT_CLASS = BalanceSheetReport
        return self.BALANCE_SHEET_REPORT_CLASS

    def get_income_statement_report_class(self):
        if not self.INCOME_STATEMENT_REPORT_CLASS:
            from django_ledger.report.income_statement import IncomeStatementReport
            self.INCOME_STATEMENT_REPORT_CLASS = IncomeStatementReport
        return self.INCOME_STATEMENT_REPORT_CLASS

    def get_cash_flow_statement_report_class(self):
        if not self.CASH_FLOW_STATEMENT_REPORT_CLASS:
            from django_ledger.report.cash_flow_statement import CashFlowStatementReport
            self.CASH_FLOW_STATEMENT_REPORT_CLASS = CashFlowStatementReport
        return self.CASH_FLOW_STATEMENT_REPORT_CLASS


lazy_loader = LazyLoader()
